# Generated by Django 2.2.20 on 2022-06-08 20:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edge', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='chunk',
            name='ref_end_index',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='chunk',
            name='ref_start_index',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]
